# :coding: utf-8

from __future__ import absolute_import
import collections
import copy
import getpass
import logging
import logging.config
import os
import tempfile

import coloredlogs

import wiz.config
import wiz.utility
import wiz.filesystem


# Configure custom colors for messages displayed in the console.
coloredlogs.DEFAULT_LEVEL_STYLES = {
    "info": {"color": "cyan"},
    "error": {"color": "red"},
    "critical": {"color": "red"},
    "warning": {"color": "yellow"}
}

#: Available levels with corresponding labels.
LEVEL_MAPPING = collections.OrderedDict([
    ("debug", logging.DEBUG),
    ("info", logging.INFO),
    ("warning", logging.WARNING),
    ("error", logging.ERROR),
])

#: Output path for files exported by default 'file' handler.
PATH = os.path.join(tempfile.gettempdir(), "wiz", "logs")

#: Default configuration for logger.
DEFAULT_CONFIG = {
    "version": 1,
    "root": {
        "handlers": ["console", "file"],
        "level": logging.DEBUG
    },
    "formatters": {
        "standard": {
            "class": "coloredlogs.ColoredFormatter",
            "format": "%(message)s"
        },
        "detailed": {
            "class": "logging.Formatter",
            "format": "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "formatter": "standard",
            "stream": "ext://sys.stdout",
            "level": logging.INFO
        },
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "detailed",
            "level": logging.INFO,
            "filename": os.path.join(PATH, "{}.log".format(getpass.getuser())),
            "maxBytes": 10485760,
            "backupCount": 20,
        }
    }
}


def initiate(console_level="info"):
    """Initiate logger configuration.

    :param console_level: Initialize the logging level for the console handler
        if possible. Default is "info".

    .. seealso:: :ref:`configuration/logging`

    """
    config = wiz.config.fetch()

    # Ensure that umask is set to 0 to create log folder with opened
    # permissions for group and other.
    _umask = os.umask(0)

    # Ensure that default output path exists.
    wiz.filesystem.ensure_directory(PATH)

    # Restore previous umask.
    os.umask(_umask)

    # Update default logging configuration if necessary.
    logging_config = copy.deepcopy(DEFAULT_CONFIG)
    wiz.utility.deep_update(logging_config, config.get("logging", {}))

    # Initiate the default level for the console if applicable.
    console_handler = logging_config.get("handlers", {}).get("console")
    if console_handler is not None:
        console_handler["level"] = LEVEL_MAPPING[console_level]

    logging.config.dictConfig(logging_config)


def capture_logs(error_stream, warning_stream):
    """Initialize logger to capture error and warning level.

    :param error_stream: instances of :class:`io.StringIO` which will receive
        all errors logged.

    :param warning_stream: instances of :class:`io.StringIO` which will receive
        all warnings logged.

    """
    logging.config.dictConfig({
        "version": 1,
        "root": {
            "handlers": ["error", "warning"],
            "level": logging.WARNING
        },
        "formatters": {
            "standard": {
                "class": "logging.Formatter",
                "format": "%(message)s"
            },
        },
        "handlers": {
            "error": {
                "class": "logging.StreamHandler",
                "formatter": "standard",
                "stream": error_stream,
                "level": logging.ERROR,
            },
            "warning": {
                "class": "logging.StreamHandler",
                "formatter": "standard",
                "stream": warning_stream,
                "level": logging.WARNING,
                "filters": ["warning-filter"]
            }
        },
        "filters": {
            "warning-filter": {
                "()": _WarningLevelFilter,
            }
        },

    })


class _WarningLevelFilter(logging.Filter):
    """Filter out log record with level other than 'WARNING'."""

    def filter(self, record):
        """Return whether *record* is a warning."""
        return record.levelno == logging.WARNING



# :coding: utf-8

import os
import getpass
import logging
import tempfile

import pytest
from six.moves import reload_module

import wiz.logging


@pytest.fixture(autouse=True)
def mock_getuser(mocker):
    """Mock getpass.getuser function."""
    mocker.patch.object(getpass, "getuser", return_value="__USER__")


@pytest.fixture()
def mocked_gettempdir(mocker):
    """Return mocked tempfile.gettempdir function."""
    return mocker.patch.object(tempfile, "gettempdir")


@pytest.fixture()
def mocked_config_fetch(mocker):
    """Return mocked wiz.config.fetch function."""
    return mocker.patch.object(wiz.config, "fetch")


@pytest.fixture()
def mocked_dict_config(mocker):
    """Return mocked logging.config.dictConfig function."""
    return mocker.patch.object(logging.config, "dictConfig")


@pytest.mark.parametrize("options, level", [
    ({}, logging.INFO),
    ({"console_level": "info"}, logging.INFO),
    ({"console_level": "debug"}, logging.DEBUG),
    ({"console_level": "warning"}, logging.WARNING),
    ({"console_level": "error"}, logging.ERROR),
], ids=[
    "simple",
    "info",
    "debug",
    "warning",
    "error",
])
def test_initiate(
    temporary_directory, mocked_gettempdir, mocked_config_fetch,
    mocked_dict_config, options, level
):
    """Initiate logger configuration."""
    mocked_gettempdir.return_value = temporary_directory
    mocked_config_fetch.return_value = {}
    reload_module(wiz.logging)

    assert not os.path.isdir(wiz.logging.PATH)

    wiz.logging.initiate(**options)

    assert os.path.isdir(wiz.logging.PATH)
    assert oct(os.stat(wiz.logging.PATH).st_mode) == oct(0o40777)

    mocked_dict_config.assert_called_once_with({
        "version": 1,
        "root": {
            "handlers": ["console", "file"],
            "level": logging.DEBUG
        },
        "formatters": {
            "standard": {
                "class": "coloredlogs.ColoredFormatter",
                "format": "%(message)s"
            },
            "detailed": {
                "class": "logging.Formatter",
                "format": "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "standard",
                "stream": "ext://sys.stdout",
                "level": level
            },
            "file": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "detailed",
                "level": logging.INFO,
                "filename": os.path.join(
                    temporary_directory, "wiz", "logs", "__USER__.log"
                ),
                "maxBytes": 10485760,
                "backupCount": 20,
            }
        }
    })


def test_initiate_with_config(
    temporary_directory, mocked_gettempdir, mocked_config_fetch,
    mocked_dict_config
):
    """Initiate logger configuration with custom config."""
    mocked_gettempdir.return_value = temporary_directory
    mocked_config_fetch.return_value = {
        "logging": {
            "root": {
                "handlers": ["console"]
            },
            "formatters": {
                "standard": {
                    "format": "%(asctime)s - %(message)s"
                }
            }
        }
    }
    reload_module(wiz.logging)

    assert not os.path.isdir(wiz.logging.PATH)

    wiz.logging.initiate()

    assert os.path.isdir(wiz.logging.PATH)
    assert oct(os.stat(wiz.logging.PATH).st_mode) == oct(0o40777)

    mocked_dict_config.assert_called_once_with({
        "version": 1,
        "root": {
            "handlers": ["console"],
            "level": logging.DEBUG
        },
        "formatters": {
            "standard": {
                "class": "coloredlogs.ColoredFormatter",
                "format": "%(asctime)s - %(message)s"
            },
            "detailed": {
                "class": "logging.Formatter",
                "format": "%(asctime)s - %(levelname)s - %(name)s - %(message)s"
            }
        },
        "handlers": {
            "console": {
                "class": "logging.StreamHandler",
                "formatter": "standard",
                "stream": "ext://sys.stdout",
                "level": logging.INFO
            },
            "file": {
                "class": "logging.handlers.RotatingFileHandler",
                "formatter": "detailed",
                "level": logging.INFO,
                "filename": os.path.join(
                    temporary_directory, "wiz", "logs", "__USER__.log"
                ),
                "maxBytes": 10485760,
                "backupCount": 20,
            }
        }
    })


def test_initiate_with_existing_folder(
    temporary_directory, mocked_gettempdir, mocked_config_fetch
):
    """Initiate logger configuration with existing folder."""
    path = os.path.join(temporary_directory, "wiz", "logs")
    os.makedirs(path)

    mocked_gettempdir.return_value = temporary_directory
    mocked_config_fetch.return_value = {}
    reload_module(wiz.logging)

    assert os.path.isdir(wiz.logging.PATH)
    assert oct(os.stat(wiz.logging.PATH).st_mode) == oct(0o40777)

    wiz.logging.initiate()

    assert os.path.isdir(wiz.logging.PATH)
    assert oct(os.stat(wiz.logging.PATH).st_mode) == oct(0o40777)


def test_initiate_with_restrictive_umask(
    temporary_directory, mocked_gettempdir, mocked_config_fetch
):
    """Initiate logger configuration when umask is restrictive."""
    os.umask(0o022)

    mocked_gettempdir.return_value = temporary_directory
    mocked_config_fetch.return_value = {}
    reload_module(wiz.logging)

    wiz.logging.initiate()

    assert os.path.isdir(wiz.logging.PATH)
    assert oct(os.stat(wiz.logging.PATH).st_mode) == oct(0o40777)
